#ifndef __KERNSH_H
#define __KERNSH_H

/* For memory limitations */
#ifndef __KERNEL__
#define __KERNEL__
#define KER 
#endif
#include <asm/page.h>
#ifdef KER
#undef __KERNEL__
#endif


#include <linux/sys.h>
#include <stdio.h>
#include <linux/module.h>

#define VERT "\033[32m"
#define NORM "\033[0m"
#define PROMPT  VERT"%"NORM" "
#define KMEM "/dev/kmem"

#ifndef NR_syscalls
#define NR_syscalls 256
#endif

#ifndef EM_486
#define EM_486 6
#endif

#ifndef PAGE_OFFSET
/* Comment this if you have an error and are sure of the value */
#error YOU MUST BE SURE OF THIS !!! PAGE_OFFSET is not defined (Please see src/include/kernsh.h)
#define PAGE_OFFSET 0xc0000000 /* Change this for UML like kernels */
#endif

#define CHECK_NUMBER_SYSCALL( n ) { \
	if ( n < 0 || n > NR_syscalls-1 ) { \
		fatal( "syscall # (%i) not in range\n", n ); \
		return; \
	} \
}

/* Don't use the old API of libdisasm */
#define LIBDISASM_NO_COMPAT 1

/* Number of slot for the hooks (in lkh.c and mkh.c) */
#define SLOT_MAX    16

extern int file_descriptor_kmem;					/* file descriptor of /dev/kmem */
extern size_t p_sys_call_table;						/* sys_call_table address */
extern size_t p_system_call;						/* system_call address */
extern unsigned long sys_call_table[ NR_syscalls ];	/* table of syscalls */
extern int modules_not_supported;					/* modules supported or not */
extern char * script_on_command_line;				/* script with option -c */
extern char * path_system_map;						/* path of system.map */
extern char * path_vmlinux;							/* path of vmlinux */
extern int mmap_use;								/* using mmap function or not */
extern int lkm_use;									/* using lkm functions or not */
extern size_t old_stub_read;
extern size_t old_stub_write;
extern int force_flag;								/* If we use the forced mode */

#define XSTRTOUL( string, addr ) { \
	addr = strtoul( string, NULL, 16 ); \
	if ( addr < 0 ) { \
		fatal( "Please, a valid address..\n\n" ); \
		return; \
	} \
}
	
#define NOARGUMENTS( n, pargc, name ) { \
	if ( pargc == n ) { \
		fatal( "No arguments presents. Type \"help\" followed by command name for\nfull documentation.\n" ); \
		return; \
	} \
}

int query_module( const char *, int, void *, size_t, size_t * );
int get_kernel_syms( struct kernel_sym * );
void * memmem( const void *, size_t, const void *, size_t );

void do_module( int argc, char * argv[] );
void do_quit( int argc, char * argv[] );
void do_symbol( int argc, char * argv[] );
void do_kmem( int argc, char * argv[] );
void do_idt( int argc, char * argv[] );
void do_gdt( int argc, char * argv[] );
void do_display_system_info( int argc, char * argv[] );
void do_bzimage( int argc, char * argv[] );
void do_virtm( int argc, char * argv[] );
void do_cap( int argc, char * argv[] );
void do_env( int argc, char * argv [] );
void do_check( int argc, char * argv [] );
void do_process( int argc, char * argv [] );
void do_sd( int argc, char * argv [] );
void do_si( int argc, char * argv [] );
void do_sh( int argc, char * argv [] );
void do_sl( int argc, char * argv [] );
void do_sr( int argc, char * argv [] );
void do_sc( int argc, char * argv [] );
void do_lhook( int argc, char * argv [] );
void do_mhook( int argc, char * argv [] );

/* usage */
void usage_symbol( void );
void usage_syscall( void );
void usage_sd( void );
void usage_si( void );
void usage_sh( void );
void usage_sl( void );
void usage_sr( void );
void usage_sc( void );
void usage_module( void );
void usage_kmem( void );
void usage_idt( void );
void usage_gdt( void );
void usage_display_system_info( void );
void usage_bzimage( void );
void usage_virtm( void );
void usage_cap( void );
void usage_env( void );
void usage_check( void );
void usage_process( void );
void usage_hook( void );
void usage_lhook( void );
void usage_mhook( void );

typedef struct command_s {
	char * name;
	void (* func)( int argc, char * argv[] );
	char * description;
	void (* usage)( void );
} command_t;

/* to add a new command, enter it in this table */
static const command_t p_command_main[] =
{
	{ "display", do_display_system_info, "System information", 
	usage_display_system_info },
	{ "symbol", do_symbol, "Handle symbols", usage_symbol },
	{ "syscall", 0, "Handle syscalls", usage_syscall },
	{ "sd", do_sd, NULL, usage_sd },
	{ "si", do_si, NULL, usage_si },
	{ "sh", do_sh, NULL, usage_sh },
	{ "sl", do_sl, NULL, usage_sl },
	{ "sr", do_sr, NULL, usage_sr },
	{ "sc", do_sc, NULL, usage_sc },
	{ "module", do_module, "Management of the modules", usage_module },
	{ "kmem", do_kmem, "Use of /dev/kmem", usage_kmem },
	{ "idt", do_idt, "Handle Interrupt Description Table", usage_idt },
	{ "gdt", do_gdt, "Handle Global Description Table", usage_gdt },
	{ "bzimage", do_bzimage, "Handle compressed kernel image", usage_bzimage },
	{ "virtm", do_virtm, "Handle virtual memory", usage_virtm },
	{ "proc", do_process, "Handle process", usage_process },
	{ "cap", do_cap, "Handle POSIX capabilities", usage_cap },
	{ "check", do_check, "Various checks against rootkits", usage_check },
	{ "hook", 0, "Playing with hooks", usage_hook },
	{ "lhook", do_lhook, NULL, usage_lhook },
	{ "mhook", do_mhook, NULL, usage_mhook },
	{ "env", do_env, "Set environnement variables", usage_env },
	{ "help", 0, "Display this help", 0 },
	{ "quit", do_quit, "Quit kernsh", 0 },
	{ NULL, 0, NULL, 0 }
};

/* kernsh */
void print_kernsh_banner( void ); 

/* memory.c */
int mmap_write_kmem( void *, size_t, int, int );
int mmap_read_kmem( void *, size_t, int, int );
int open_kmem( char * );
void close_kmem( int );
int write_kmem( void *, size_t, int, int );
int read_kmem( void *, size_t, int, int );
void kdump( size_t, int, int );
void kdisass( size_t, int, int );
void kwrite( size_t, char * );
void kscan( size_t, size_t, char *, size_t, int, int );
int create_rw_km( int );
void toggle_lkm_use( void );
void kkmalloc(int , unsigned long * );
void kkfree(size_t);
#define NO_JOKER 32768 /* > 256 */
/* choose the file desciptor */
size_t find_area( size_t, size_t, void *, int, int, int ); 
/* wrapper for file_descriptor_kmem */
size_t kfind_area( size_t, size_t, void *, int, int );
void kinsmod( char *, char * );

/* symbol.c */
size_t get_symbol_by_kernel_sym( char * );
size_t get_symbol_by_query_module( char * );
size_t get_symbol_by_system_map( char *, int );
size_t get_symbol_by_name( char * );
char * get_symbol_by_addr( char * );
size_t get_next_symbol_by_addr( size_t );
int found_vmlinux( void );
FILE * found_system_map( void );
int my_strncmp( char * , char * , int );
int foundnext( char * , char , int );

/* misc.c */
char * dupstr( char * );
void fatal( const char *, ... );
void hexdump( unsigned char *, unsigned int, size_t );
int parse( char *, char *, size_t, int );
char enter( void );
char * stripwhite( char * );
char * string2hex( char *, int * );
void * ksh_xmalloc( size_t );
int file_is_present( char * );
void *memmem_j(const void *, size_t , const void *, size_t , const char );

/* idt.c */
struct s_idtr {
    unsigned short limit;
    unsigned int base;
} __attribute__ ((packed))idtr;

struct {
    unsigned short off1;
    unsigned short sel;
    unsigned char none;
	unsigned char flags;
    unsigned short off2;
} __attribute__ ((packed))idt;

unsigned char * get_idt_segment( unsigned int );

/* gdt.c */
struct s_gdtr {
unsigned short limit;
unsigned int base;
} __attribute__ ((packed))gdtr;

/* syscall.c */
int set_syscall_table( void );
int set_system_call( void );
int refresh_syscall_table( void );
int scall( int num, int argc, unsigned int argv[] );

/* SYSCALL MACRO - Kstat (hum no) It's better to use scall than 
 * these definitions
 */
#define SYSCALL0(num) asm volatile ( "\n\tint $0x80" : "=a"(__ret) : "0"(num) );
#define SYSCALL1(num,arg0) asm volatile ( "\n\tint $0x80": "=a"(__ret) : "0"(num), "b" (arg0) );
#define SYSCALL2(num,arg0,arg1) asm volatile ( "\n\tint $0x80": "=a"(__ret) : "0"(num),"b" (arg0),"c"(arg1) );
#define SYSCALL3(num,arg0,arg1,arg2) asm volatile ( "\n\tint $0x80": "=a"(__ret) : "0"(num),"b" (arg0),"c"(arg1),"d"(arg2) );
#define SYSCALL4(num,arg0,arg1,arg2,arg3) asm volatile ( "\n\tint $0x80": "=a"(__ret) : "0"(num),"b" (arg0),"c"(arg1),"d"(arg2),"S"(arg3) );
#define SYSCALL5(num,arg0,arg1,arg2,arg3,arg4) asm volatile ( "\n\tint $0x80": "=a"(__ret) : "0"(num),"b" (arg0),"c"(arg1),"d"(arg2),"S"(arg3),"D"(arg4) );
/* 6 args are a lot tricky so prefer to pass a pointer to the arguments (argc, argv like)*/

/* md5.c */
struct MD5Context {
	unsigned long buf[ 4 ];
	unsigned long bits[ 2 ];
	unsigned char in[ 64 ];
};

void MD5Init( struct MD5Context * context );
void MD5Update( struct MD5Context * context, unsigned char const * buf,
    unsigned len );
void MD5Final( unsigned char digest[ 16 ],
    struct MD5Context * context );
void MD5Transform( unsigned long buf[ 4 ],
    const unsigned char in[ 64 ] );
void dumpmd5( char * pass, int value,
    unsigned char md5buffer[ BUFSIZ ] );

/* signature.c */
#define SIG_FOR_SYSCALL 0
#define SIG_FOR_IDT 1

void create_file_signatures( char *, char *, int, int );
int check_file_signatures( char *, char *, int, int );

/* disasm.c */
void display_code_asm( size_t, unsigned char *, int );
int make_esp( unsigned char *, size_t , unsigned char *, size_t * );
int disasm( size_t, int );
extern int disasm_syn;

/* virtm.c */
struct mem_addr {
	unsigned long vm_start;
	unsigned long vm_end;
	unsigned long vm_start_cache;
	unsigned long vm_end_cache;
	int map_count;
	unsigned long startcode;
	unsigned long endcode;
	unsigned long startdata;
	unsigned long enddata;
	unsigned long startbrk;
	unsigned long startstack;
	unsigned long arg_start;
	unsigned long arg_end;
	unsigned long env_start;
	unsigned long env_end;
	unsigned long rss;
	unsigned long total_vm;
	unsigned long locked_vm;
	unsigned long swap_address;
};

extern int parse_getvirtaddr( struct mem_addr *, int );
extern int parse_rwvirtaddr( int, char *, int, size_t, int );

/* env.c */
char * parse_line( char * );
void end_parse_line( void );
#define VAR_SIZE 16
struct Env {
	char name[VAR_SIZE];
	struct Env * next;
	char * value;
};
void clear_all(struct Env *);
extern struct Env env;
extern FILE * old_std;

/* check_op.c */
int check_my_file_operation( size_t );
int check_my_file_operation_by_name( char * );

/* module.c */
void get_info_module( char *, size_t );

/* process.c */
int is_hided( int );

/* ts.c */
void proc_kps( void );
void proc_khide_detect( void );
void proc_khide( int );
void proc_kts( int );
void proc_kfile( int );
void proc_kinfo( int );

/* display.c */
int my_fprintf (FILE *stream, const char *format, ...);
extern int be_paranoid;
/* automatic substitution to avoid people getting confused ;) */
#define fprintf my_fprintf 

/* usage.c */
typedef struct usage_s {
	char * command_line;
	char * description;
	char * arguments;
} usage_t;

void usage( usage_t * );

#endif /* __KERNSH_H */

