/*
 * The project kernsh is (C) 2002-2004. This program is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * Version 2. This guarantees your right to use, modify, and
 * redistribute this software under certain conditions.
 *
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#include <linux/sys.h>
#include <sys/utsname.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
/* Signal Handling */
#include <signal.h>
/* Error handling */
#include <setjmp.h>

#include "include/kernsh.h"
#include <readline/readline.h>
#include <readline/history.h>

#define VERSION "0.2c"
#define HIST_SAV	500
#define HIST_FILE	".kernsh_history"

/* Silly helper */
#define FREE(p) if ( p != NULL ) free(p) ;
#define FREE2(p,d) if ( p != NULL && p != d ) free(p);

size_t p_sys_call_table;
size_t p_system_call;
unsigned long sys_call_table[ NR_syscalls ];
int file_descriptor_kmem;
int modules_not_supported = 0;
char * script_on_command_line = NULL;
char * path_system_map = NULL;
char * path_vmlinux = NULL;
char * path_file_kmem = NULL;
char * path_file_hist = NULL; /* History file */
int created;		/* know if we create the /dev/kmem equivalent */
int mmap_use = 0;	/* define if we want to use mmap read/write functions */
int lkm_use = 0;	/* define if we want to use lkm read/write functions */
int force_flag = 0;	/* Use this if you kernsh failed to be launch ... */

static jmp_buf ex_env; /* When an Oops occurs be able to save ourselve */
static int env_safe = 0; /* Test if we save something or not */
static int in_readline = 0; /* Test if we are in readline state */

void
print_kernsh_banner( void )
{
	fprintf( stdout, "-----------------------[ KERNSH %s (c) 2004 ]-----------------------\n\n",
		VERSION );
}

void
print_kernsh_usage( char * kernsh )
{
    fprintf( stderr, "[!] Usage: %s [options] <arguments>..\n",
        kernsh );
    fputs( "[!]\n", stderr );

    fputs( "[!] __options\n", stderr );
    fputs( "[!]    -h : this help\n", stderr );
    fputs( "[!]    -q : mode quiet\n", stderr );
    fputs( "[!]    -c <file> : submit a script to kernsh on "
        "command line\n", stderr );
    fputs( "[!]    -s <file> : path of System.map\n", stderr );
    fputs( "[!]    -v <file> : path of vmlinux\n", stderr );
    fputs( "[!]    -m : define if we want to use mmap read/write functions\n", 
		stderr );
	fputs( "[!]    -l : define if we want to use lkm read/write functions\n", 
		stderr );
    fputs( "[!]    -k <file> : replace /dev/kmem by your proper file\n", stderr );
	fputs( "[!]    -f : force the use of the some symbols from System.map\n", stderr );
	fputs( "[!]    -j <file> : use this as history file. -1 disable it\n"
			,stderr);
	fprintf(stdout, "[!]    (The default is %s/"HIST_FILE" or $KERNSH_HOME/"HIST_FILE")\n"
			,getenv("HOME"));
    fputs( "[!]\n\n", stderr );
    exit( EXIT_FAILURE );
}

/* generator function for command completion.  STATE lets us know 
 * whether to start from scratch; without any state (i.e. STATE == 0), 
 * then we start at the top of the list. 
 * http://www.ugcs.caltech.edu/info/readline/rlman_2.html
 */
char *
command_generator( char * text, int state )
{
    char * name = NULL;
	static int list_index;
	static int len;

	/* if this is a new word to complete, initialize now.  This 
	 * includes saving the length of TEXT for efficiency, and 
	 * initializing the index variable to 0. 
	 */
	if ( !state ) {
		list_index = 0;
		len = strlen( text );
	}

	/* return the next name which partially matches from the 
	 * command list. 
	 */
	while ( (name = p_command_main[list_index].name) ) {
		list_index++;
		if ( strncmp(name, text, len) == 0 ) {
			return( dupstr(name) );
		}
	}

	/* if no names matched, then return NULL. */
	return( (char *)NULL );
}

/* attempt to complete on the contents of TEXT.  START and END show the
 * region of TEXT that contains the word to complete.  We can use the
 * entire line in case we want to do some simple parsing.  Return the
 * array of matches, or NULL if there aren't any. 
 * http://www.ugcs.caltech.edu/info/readline/rlman_2.html
 */
char **
fileman_completion( char * text, int start, int end )
{
	char ** matches = NULL;

	/* If this word is at the start of the line, then it is a command
	 * to complete.  Otherwise it is the name of a file in the current
	 * directory. 
	 */
	if ( start == 0 ) {
		matches = rl_completion_matches( text, 
			(rl_compentry_func_t *)command_generator );
	}

	return( matches );
}

void 
initialize_readline( void )
{
	/* allow conditional parsing of the ~/.inputrc file. */
	rl_readline_name = "kernsh";

	/* tell the completer that we want a crack first. */
	rl_attempted_completion_function = 
		(CPPFunction *)fileman_completion;

	/* Initialize the history */
	if ( path_file_hist[0] != '-' 
			&& read_history(path_file_hist) != 0 ) {
		/* Not history file doesn't matter */
	}
		
	return;
}

void
end_session( void )
{
	int i;
	
	fputs( "\n", stdout );
	if ( lkm_use == 1 )
	{
		toggle_lkm_use( );
	}

	close_kmem( file_descriptor_kmem );
	
	if ( created != 0 ) {
		i = unlink( path_file_kmem );
		if ( i ) {
			perror( "end_session:unlink()" );
		}
	}

	clear_all(&env); /* clear the environnement */

	/* Write history file before clearing it*/
	if ( path_file_hist[0] != '-' ) {
		if( append_history(HIST_SAV,path_file_hist) != 0 ) {
			write_history(path_file_hist);
		}
	}
	clear_history(); /* Clear Readline history */
	
	rl_discard_keymap( rl_get_keymap() ); /* Clear the keymap */
	rl_set_prompt(NULL); /* Clear the prompt */
	/* Freeing pathnames */
	FREE(path_file_hist);
	FREE(path_vmlinux);
	FREE(path_system_map);
	FREE2(path_file_kmem,KMEM);

	exit( EXIT_SUCCESS );
}

void 
do_quit( int argc, char * argv[] )
{
	end_session();
}

#define ISBLANK( ch ) ( ((ch) == ' ') || ((ch) == '\t') )
int 
command( char * cmd )
{
	char command_name[ BUFSIZ ];
	char * p_command_name = NULL;
	char * argv[64];
	char * string = NULL;
	static const char delim[] = " \t\n";
	int argc = 0;
	int n;
	int i;
	int k;

	p_command_name = cmd;
	while ( *p_command_name && !ISBLANK(*p_command_name) ) {
		p_command_name++;
	}

	memcpy( command_name, cmd, p_command_name-cmd );
	command_name[p_command_name - cmd] = '\0';
	
	memset( argv, '\0', sizeof(argv) );
	argv[argc++] = p_command_name;
	string = (char *)strtok( p_command_name, delim );
	while ( string != NULL ) {
		argv[argc++] = string;
		string = (char *)strtok( NULL, delim );
	}
	argv[argc] = '\0';

	for ( i = 0; i < sizeof(p_command_main)-1; i++ ) {
		if ( p_command_main[i].name == NULL ) {
			break;
		}

		n = strcmp( p_command_main[i].name, command_name );
		if ( n == 0 ) {
			break;
		}
	}

	if ( p_command_main[i].name == "help" ){
		if ( argc == 1 ) {
			for ( i = 0; i < sizeof(p_command_main)-1; i++ ) {
        		if ( p_command_main[i].name == NULL ) {
					break;
				}

				if ( (p_command_main[i].name != "module" 
				|| (p_command_main[i].name == "module" && !modules_not_supported)) 
				&& p_command_main[i].description != NULL ) {
					fprintf( stdout, "%s\t: %s\n", 
					p_command_main[i].name, 
					p_command_main[i].description );
				}
			}
			fputs( "\nType \"help\" followed by command name "
			"for full documentation.\n\n", stdout );
			return( 0 );
		} else if ( argc == 2 ) {
			for ( k = 0; k < sizeof(p_command_main)-1; 
				k++ ) {
				if ( p_command_main[k].name == NULL  ) {
					fatal( "undefined command..\n\n" );
					return( -1 );
				}

				if( !strcmp(p_command_main[k].name, argv[1])  ) {
					if(  p_command_main[k].usage != 0 ) {
						p_command_main[k].usage(); 
						return( 0 );
					} else {
						fatal( "no avaible help for this command\n\n" );
						return( -1 );
					}
				} else if ( !strcmp(p_command_main[k].name, argv[1]) ) {
					if ( p_command_main[k].usage != 0 ) {
						p_command_main[k].usage();
						return( 0 );
					}
				}
			}
		} else {
			fatal( "usage\n" );
			return( 0 );
		}
	}

	if ( p_command_main[i].name == NULL ) {
		fatal( "undefined command..\n\n" );
		return( -1 );
	}

	/* This is not so nice, but ... */
	if ( 0 == strncmp(p_command_main[i].name,"quit",sizeof("quit")-1))
	{
		free(cmd);
	}
	
	if ( !p_command_main[i].func ) {
		fprintf( stdout, "Type \"help\" followed by this command name for full "
			"documentation.\n" );
	} else {
		p_command_main[i].func( argc, argv );
	}
	fputs( "\n", stdout );
	return( 0 );
}

void
initialization( int quiet )
{
	FILE * fd_file_system_map;
	struct utsname p_uname;
	struct stat file;
	int i;
	int n;

	if ( stat(path_file_kmem, &file) == 0 ) {
		created = 0;
	} else if ( errno == ENOENT ) {
		i = mknod( path_file_kmem, S_IFCHR|S_IRUSR|S_IWUSR, 256+2 );
		if ( i == -1 ) {
			if ( !quiet ) {
				perror( "initialization:mknod()" );
			}
			fatal( "can't create the special device (%s)..\n", path_file_kmem );
			exit( EXIT_FAILURE );
		} else {
			created = 1;
		}
	}

	if ( !quiet ) {
		uname( &p_uname );
		fprintf( stdout, "system name  : %s\n", p_uname.sysname );
		fprintf( stdout, "node name    : %s\n", p_uname.nodename );
		fprintf( stdout, "realease     : %s\n", p_uname.release );
		fprintf( stdout, "version      : %s\n", p_uname.version );
		fprintf( stdout, "machine name : %s\n", p_uname.machine );
		fprintf( stdout, "domain name  : %s\n", p_uname.__domainname );
		fprintf( stdout, "using mmap   : %s\n", mmap_use ? "yes" : "no" );
		fprintf( stdout, "using lkm    : %s\n\n", lkm_use ? "yes" : "no" );
	}

	/* open /dev/kmem */
	file_descriptor_kmem = open_kmem( path_file_kmem );
	if ( file_descriptor_kmem < 0 ) {
		fatal( "can't open /dev/kmem..\n" );
		exit( -1 );
	}

	/* now the file descriptor is open so we can delete it ;) */
	if ( created != 0 ) {
		if ( unlink(path_file_kmem) == -1 ) {
			fatal( "can't remove %s, please do it by hand\n", path_file_kmem );
		} else {
			created = 0;
		}
	}

	/* check if System.map is present */
	fd_file_system_map = found_system_map();
	if ( fd_file_system_map == NULL ) {
		fatal( "[#] System.map is necessary to kernsh.\n"
			"[#] Run the script scripts/create_system_map.sh to create it and copy it in /boot\n"
			"[#] Run again kernsh\n\n" );
		exit( EXIT_FAILURE );
	}
	fclose( fd_file_system_map );

	/* we're checking integrity of system_call and sys_call_table 
	 * before to continue.
	 */
	n = set_system_call();
	if ( n < 0 ) exit( EXIT_FAILURE );
	n = set_syscall_table();
	if ( n < 0 ) exit( EXIT_FAILURE );
	n = refresh_syscall_table();
	if ( n < 0 ) exit( EXIT_FAILURE );

	/* Setting the default env */
	sprintf( env.name, "?" );
	env.value = malloc( BUFSIZ );
	memset( env.value, 0, BUFSIZ );

	/*  Enabling the lkm read/write  */
	if ( lkm_use == 1) {
		old_stub_read = create_rw_km( 0 );
		old_stub_write = create_rw_km( 1 );
	}
	
	return;
}	

void
sig_hand(int sign)
{
	switch(sign) {
		case SIGFPE:
		case SIGILL:
		case SIGBUS:
		case SIGSEGV:
			switch ( env_safe ) {
				case 1:
					longjmp(ex_env,1);
					break;
				case 2:
					fatal( "Hum still a segfault ? avoid a loop \n" );
					exit( EXIT_FAILURE );
					break;
				default:
					break;
			}
			fatal( "SegFault (%d) \n",sign);
			break;
		case SIGTERM:
			fatal( "Want to Quit ?\n" );
			break;
		case SIGCHLD: /* Ignore Pipe child */
			break;
		case SIGINT: /* Ignore Ctrl-C ? */
			fatal( "quit or Ctrl-D\n" );
			break;
		default:
			fatal( "Received : %d\n",sign);
			break;
	}
	/* Restore readline handlers */
	if ( in_readline == 1 )	{
		rl_free_line_state();
		rl_cleanup_after_signal(); 
/*		rl_reset_after_signal(); */
		rl_resize_terminal();
	}
}

#define SET_HAND(x,y) if ( -1 == sigaction(x,y,NULL) ) \
		{ \
			fatal("Can't set the action for %d\n",x); \
		}

void init_sign ( void )
{
	/* Signal Handling */
	struct sigaction new,old;
	
	memset(&new,0,sizeof(new));
	new.sa_handler = &sig_hand;
	new.sa_flags |= SIG_SETMASK;
	if ( -1 == sigfillset(&new.sa_mask) ) {
		fatal( "Can't fill the sigset\n" );
		exit( EXIT_FAILURE );
	}
	if ( -1 == sigaction(SIGSEGV,&new,&old) ) {
		fatal( "Can't set the action for SIGSERV\n" );
	}
	SET_HAND(SIGHUP,&new);
	SET_HAND(SIGINT,&new);
	SET_HAND(SIGQUIT,&new);
	SET_HAND(SIGILL,&new);
	SET_HAND(SIGTRAP,&new);
	SET_HAND(SIGABRT,&new);
	SET_HAND(SIGBUS,&new);
	SET_HAND(SIGFPE,&new);
	SET_HAND(SIGUSR1,&new);
	SET_HAND(SIGUSR2,&new);
	SET_HAND(SIGPIPE,&new);
	SET_HAND(SIGALRM,&new);
	SET_HAND(SIGTERM,&new);
	SET_HAND(SIGCHLD,&new);
	SET_HAND(SIGTSTP,&new);
/*	SET_HAND(SIGTTIN,&new); */ /* Readline */
/*	SET_HAND(SIGTTOU,&new); */ /* Readline */
	SET_HAND(SIGURG,&new);
	SET_HAND(SIGXCPU,&new);
	SET_HAND(SIGXFSZ,&new);
	SET_HAND(SIGVTALRM,&new);
	SET_HAND(SIGPROF,&new);
/*	SET_HAND(SIGWINCH,&new); */ /* Readline */
	SET_HAND(SIGIO,&new);
}

int 
main( int argc, char * argv[] )
{
	FILE * fpscript = NULL;
	char * line = NULL;
	char * s = NULL;
	char buffer[ 128 ];
	char c = 0;
	uid_t uid;
	int fd;
	int quiet;
	int i;
	
	/* test uid */
	uid = getuid(); 
	if ( uid ) { 
		fatal( "Hey ! You must be root\n" );
		exit( EXIT_FAILURE );
	}
	
	init_sign(); /* init the signals */

	print_kernsh_banner();

	/* Before the script test ... */
	quiet = mmap_use = 0;
	
	if ( (argc == 2 && argv[1][0] != '\x2d') ||
			(argc == 3 && argv[2][1] == 'f' && argv[1][0] != '\x2d')) {
		if ( argc == 3 && argv[2][1] == 'f' ) {
			force_flag = 1;
			c = (char)1;
		} else {
/*		command( argv[1] );*/
/*		return( -1 );*/
			c = (char)1;
		}
		/* be  quiet by default ? yes */
		quiet = 1;
	}

	if ( c != 1 )
    while ( (c = getopt(argc, argv, "fhqmc:s:v:k:lj:")) != EOF ) {
        switch ( c ) {
			case 'j':
				path_file_hist = dupstr( optarg );
				printf("Using %s as history file\n",path_file_hist);
				break;
			case 'l':
				lkm_use = 1;
				break;
			case 'f':
				force_flag = 1;
				break;
			case 'q':
				quiet++;
				break;
			case 'c':
				script_on_command_line = dupstr( optarg );
				break;
			case 'm':
				mmap_use = 1;
				break;
            case 's':
                path_system_map = dupstr( optarg );
                break;
            case 'v':
                path_vmlinux = dupstr( optarg );
                break;
            case 'h':
                print_kernsh_usage( argv[0] );
                break;
			case 'k':
				path_file_kmem = dupstr( optarg );
				break;
            default:
                print_kernsh_usage( argv[0] );
                exit( EXIT_FAILURE );
        }
    }

	if ( path_file_kmem == NULL ) {
		path_file_kmem = KMEM;
	}

	if ( path_file_hist == NULL ) {
		char* tmp = NULL;
		if( ((tmp = getenv("KERNSH_HOME")) != NULL) 
				|| ((tmp = getenv("HOME")) != NULL )) {
			path_file_hist = strcat(realloc(dupstr(tmp),strlen(tmp)+sizeof("/"HIST_FILE)+2)
					,"/"HIST_FILE);
		}
	}

	initialization( quiet );
	initialize_readline();

	/* test if modules are supported */
#define KSYMS "/proc/modules"
	fd = open( KSYMS, O_RDONLY );
	if ( fd < 0 ) {
		fatal( "[!] Modules aren't supported..\n\n" );
		modules_not_supported++;
	}
	close( fd );
	
	if ( script_on_command_line ) {
		fpscript = fopen( script_on_command_line, "r" );
		if ( !fpscript ) {
			perror( "main:fopen()" );
			exit( EXIT_FAILURE );
		}

		for ( i = 0 ; 
			fgets(buffer, sizeof(buffer), fpscript) != NULL ; i++ ) {
			buffer[ strlen(buffer)-1 ] = '\0';

			s = stripwhite( buffer );
			if ( *s ) {
				command( buffer );
			}
		}
		fclose( fpscript );
	} else if ( c == 1 ) {
		command( argv[1] );
		exit( EXIT_SUCCESS );
	} else {
		for ( ;; ) {
			if (setjmp(ex_env) == 0) {
				in_readline = 1;
				line = readline( PROMPT );
				in_readline = 0; /* Here is a race condition :/ */
				env_safe = 1;
				if ( !line ) {
					break;
				}

				line = parse_line( line );

				/* remove leading and trailing whitespace from the line.
			 	 * then, if there is anything left, add it to the history list
			 	 * and execute it. 
				 */
				s = stripwhite( line );
				if ( *s ) {
					add_history( s );
					command( s );
				}

				end_parse_line();
				free( line );

			} else {
				env_safe = 2;
				fatal( "Oops ! restarting :)\n" );
			}
		}
	}

	end_session();
	exit( EXIT_SUCCESS );
}
