/*
 * The project kernsh is (C) 2002-2004. This program is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * Version 2. This guarantees your right to use, modify, and
 * redistribute this software under certain conditions.
 *
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

/*
 * Dictracy Loadable Kernel Module
 *                          by -  twiz   - twiz@email.it
 *                                thefly - thefly@acaro.org
 *
 * That module let you investigate, read, search, dump and write the virtual
 * address space of a process running.
 *
 * From the idea exposed by vecna in rmfbd :
 *    http://www.s0ftpj.org/bfi/dev/BFi11-dev-06
 *
 * Thanks : silvio, Liv Tyler (by thefly)
 */

#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <linux/fs.h>
#include <errno.h> // For -ENOSYS
#include "include/kernsh.h"
#include "../libdisasm/libdis.h"
#include <nrsys.h> /* Automatically created */

/* XXX Should be move to a common header with modules/lkm-virtm.o */
#define __NR_getexeimage NR_sysfree0
#define __NR_getvirtaddr NR_sysfree1
#define __NR_rwvirtaddr NR_sysfree2

int
parse_getvirtaddr( struct mem_addr * mem, int pid )
{
	int __ret;
	
	SYSCALL2( __NR_getvirtaddr, mem, pid );
	return __ret;
}

int
parse_rwvirtaddr( int pid, char * buf, int size, size_t addr, int what )
{
	int __ret;

	SYSCALL5( __NR_rwvirtaddr, pid, buf, size, addr, what );
	return __ret;
}

int
parse_getexeimage( int pid, char * filename )
{
	int __ret;

	SYSCALL2( __NR_getexeimage, pid, filename );
	return __ret;
}

void
virtm_error_message( int i )
{
	if ( i == 1 ) fatal( "is the pid correct ?\n" );
	if ( i == -1 ) fatal( "out of memory\n" );
	if ( i == -2 ) fatal( "invalid or reserved page\n" );
}

void
get_virtual_memory( pid_t pid )
{
	struct mem_addr addresses;
    int i;
	
    i = parse_getvirtaddr( &addresses, pid );
    if ( i ) {
		virtm_error_message( i );
		return;
    }

    fprintf( stdout, "[#] virtual memory for pid %d:\n\n", pid );
    fprintf( stdout, "vm_start(1)       : \t%#x\n", (size_t)addresses.vm_start );
    fprintf( stdout, "vm_end(1)         : \t%#x\n", (size_t)addresses.vm_end );
    fprintf( stdout, "vm_start_cache(1) : \t%#x\n", (size_t)addresses.vm_start_cache );
    fprintf( stdout, "vm_end_cache(1)   : \t%#x\n", (size_t)addresses.vm_end_cache );
    fprintf( stdout, "map_count         : \t%#x\n", (size_t)addresses.map_count );
    fprintf( stdout, "code_start        : \t%#x\n", (size_t)addresses.startcode );
    fprintf( stdout, "code_end          : \t%#x\n", (size_t)addresses.endcode );
    fprintf( stdout, "data_start        : \t%#x\n", (size_t)addresses.startdata );
    fprintf( stdout, "data_end          : \t%#x\n", (size_t)addresses.enddata );
    fprintf( stdout, "heap_start        : \t%#x\n", (size_t)addresses.startbrk );
    fprintf( stdout, "stack_start       : \t%#x\n", (size_t)addresses.startstack );
    fprintf( stdout, "arg_start         : \t%#x\n", (size_t)addresses.arg_start );
    fprintf( stdout, "arg_end           : \t%#x\n", (size_t)addresses.arg_end );
    fprintf( stdout, "env_start         : \t%#x\n", (size_t)addresses.env_start );
    fprintf( stdout, "env_end           : \t%#x\n", (size_t)addresses.env_end );
    fprintf( stdout, "rss               : \t%#x\n", (size_t)addresses.rss );
    fprintf( stdout, "total_vm          : \t%#x\n", (size_t)addresses.total_vm );
    fprintf( stdout, "locked_vm         : \t%#x\n", (size_t)addresses.locked_vm );
    fprintf( stdout, "swap_address      : \t%#x\n", (size_t)addresses.swap_address );
}

#define READ 0
#define WRITE 1

void
read_virtual_memory( pid_t pid, size_t addr, int ssize )
{
    char * buffer = NULL;
    int i;

    buffer = malloc( ssize );
    if ( buffer == NULL ) {
        perror( "read_virtual_memory:malloc()" );
        return;
    }

    i = parse_rwvirtaddr( pid, buffer, ssize, addr, READ );
    if ( i ) {
		virtm_error_message( i );
		free( buffer );
        return;
    }

    fprintf( stdout, "\n[ %#x -> %#x | pid : %d ]\n", addr,
        addr+ssize, pid );
    hexdump( buffer, ssize, addr );
    free( buffer );
    return;
}

void
write_virtual_memory( pid_t pid, size_t addr, char * buffer )
{
	char * str2hex = NULL;
	char c;
	int i;

	str2hex = string2hex( buffer, &i );
	if ( str2hex == NULL ) {
		return;
	}

	fputs( "\n[ before ]", stdout );
	read_virtual_memory( pid, addr, i );

	fputs( "[ after ]\n", stdout );
	hexdump( str2hex, i, addr );

	fputs( "\nreally writing [y/n] ? ", stdout );
	c = getchar();
	getchar();
	if ( c == 'y' || c == 'Y' ) {
    	i = parse_rwvirtaddr( pid, str2hex, strlen(str2hex), 
			addr, WRITE );
    	if ( i ) {
			virtm_error_message( i );
			free(str2hex);
			return;
    	}
		fprintf( stdout, "strings %s written\n", buffer );
	} else {
		fatal( "can't write on a part of virtual memory\n" );
	}

	fputs( "\n", stdout );
	free( str2hex );
	return;
}

void
search_virtual_memory( int pid, size_t lower, size_t upper, 
	char * string )
{
	size_t addr = lower;
	int len_string = strlen( string );
	char buf[ len_string ];
	int i;
	int p = 0;

	bzero( buf, len_string );

	for ( addr = lower; addr < upper; addr++ ) {
		i = parse_rwvirtaddr( pid, buf, len_string, addr, READ );
		if ( !i ) {
			if ( !strncmp(buf, string, len_string) ) {
				fprintf( stdout, "pattern found at %#x\n", addr );
				p++;
			}
		}
	}
    if ( p == 0 ) {
        fatal( "no pattern found\n" );
    } else {
        fprintf( stdout, "found %d pattern \n", p );
    }
    return;
}	
	
void
disassemble_virtual_memory( pid_t pid, size_t startaddr, int ssize )
{
    struct mem_addr memory;
	unsigned char * image = NULL;
    int i;

    i = parse_getvirtaddr( &memory, pid );
    if ( i ) {
        fatal( "can't get virtual memory for pid %d\n", pid );
        return;
    }

	image = malloc( ssize );
	if ( image == NULL ) {
		perror( "disassemble_virtual_memory:malloc()" );
		return;
	}
	
	i = parse_rwvirtaddr( pid, image, ssize, startaddr, READ );
	if ( i ) {
		virtm_error_message( i );
		free( image );
		return;
	}

	fprintf( stdout, "[!] disassembling pid %d from address %#x..\n", 
		pid, startaddr );
	fprintf( stdout, "[!] end address: %#x\n", startaddr+ssize );
	fprintf( stdout, "[!] len: %d\n\n", ssize );
	display_code_asm( startaddr, image, ssize );
	free( image ); /* Doh !!!! */
	return;
}	

void
get_exe_image( pid_t pid, char * filename )
{
	int i;
	
	fprintf( stdout, "[!] get exe image from process pid %d	..\n", pid );
	i = parse_getexeimage( pid, filename );
	if ( i ) {
		fatal( "can't get exe image ..\n" );
	}

	fprintf( stdout, "[!] see now %s exe image ..\n", filename );
	return;
}

int
check_valid_addresses( pid_t pid, size_t addr )
{
	struct mem_addr addresses;
	int i;

	i = parse_getvirtaddr( &addresses, pid );
	if ( i ) {
		virtm_error_message( i );
		return( -1 );
	}

#define INVALID "invalid address, run 'virtm -g <pid>' to get a valid address\n"
	if ( addr < addresses.vm_start ) {
		fatal( INVALID );
		return( -1 );
	}
	
	if ( addr > addresses.vm_end ) {
		if ( addr < addresses.startdata ) {
			fatal( INVALID );
			return( -1 );
		}
	}
	
	if ( addr > addresses.enddata ) {
		if ( addr < addresses.vm_start_cache ) {
			fatal( INVALID );
			return( -1 );
		}
	}
	
	if ( addr > addresses.vm_end_cache ) {
		fatal( INVALID );
		return( -1 );
	}

	return( 0 );
}

void
usage_virtm( void )
{
    usage_t p_usage;

    p_usage.command_line =
        "virtm [ -sgrwdef ] <arguments>";
    p_usage.description =
 		"This command makes it possible to read, search, dump and write the virtual\n"
 		"address space of a process running.";
    p_usage.arguments = 
		"  -g [pid] : get virtual addresses on the queried process ;\n"
		"  -r [pid] [start address] [size] : read a part of virtual memory ;\n"
		"  -w [pid] [address] [string] : write on a part of virtual memory ;\n"
		"  -s [pid] [start address] [end address] [string] : search a particular\n" 
		"virtual memory area ;\n"
		"  -d [pid] [start address] [size] : disassemble the queried process ;\n"
		"  -e [pid] [filename] : get exe image of a [pid] process from virtual memory and "
		"write image to [filename] ;\n"
		"  -f : force to bypass control for memory boundaries: Do it and pray :)\n"
		"Could be deadly with the -w option (and this is not a joke).";

    usage( &p_usage );
    return;
}

enum {
	READ_VIRTUAL_MEMORY,
	WRITE_VIRTUAL_MEMORY,
	SEARCH_VIRTUAL_MEMORY,
	DISASS_VIRTUAL_MEMORY,
	EXE_VIRTUAL_MEMORY,
};

void
do_virtm( int pargc, char * pargv[] )
{
	pid_t pid;
	char c;
	size_t l;
	size_t ll;
	int flag,force;
	int i;

	if ( parse_rwvirtaddr( 0, NULL, 0, 0, 0 ) == -ENOSYS) {
		fatal( "module lkm-virtm not loaded\n");
		return;
	}
	
	force = 0;
	flag = -1;
	optind = 0;
	
	NOARGUMENTS( pargc, 1, virtm );

	while ( (c = getopt(pargc, pargv, "fwrseg:d:")) != EOF ) {
		switch ( c ) {
			case 'f':
				force = 1;
				break;
			case 'g': 
				get_virtual_memory( atoi(optarg) );
				break;
			case 'r':
				flag = READ_VIRTUAL_MEMORY;
				break;
			case 'w':
				flag = WRITE_VIRTUAL_MEMORY;
				break;
			case 's':
				flag = SEARCH_VIRTUAL_MEMORY;
				break;
			case 'd':
				flag = DISASS_VIRTUAL_MEMORY;
				break;
			case 'e':
				flag = EXE_VIRTUAL_MEMORY;
				break;
			default:
				usage_virtm();
				return;
		}
	}

	/* Shit :) */
	if ( force ) {
		pargc--;
	}
	
	switch ( flag ) {
	case READ_VIRTUAL_MEMORY:
		if ( pargc != 5 ) {
			usage_virtm();
			return;
		}

		pid = atoi( pargv[optind] );
		XSTRTOUL( pargv[optind+1], l );
		if ( !force ) {
			i = check_valid_addresses( pid, l );
			if ( i ) {
				return;
			}
		}

		read_virtual_memory( pid, l, atoi(pargv[optind+2]) );
		break;

	case WRITE_VIRTUAL_MEMORY:
		if ( pargc != 5 ) {
			usage_virtm();
			return;
		}

		pid = atoi( pargv[optind] );
		XSTRTOUL( pargv[optind+1], l );
		
		if ( !force ) {
			i = check_valid_addresses( pid, l );
			if ( i ) {
				return;
			}
		}

		write_virtual_memory( pid, l, pargv[optind+2] );
		break;

	case SEARCH_VIRTUAL_MEMORY:
		if ( pargc != 6 ) {
			usage_virtm();
			return;
		}

		pid = atoi( pargv[optind] );
		XSTRTOUL( pargv[optind+1], l );
		XSTRTOUL( pargv[optind+2], ll );
        if ( ll - l < 0 ) {
            fatal( "size of memory area you want to scan "
                "is negative !\n" );
            return;
        }

		if ( !force ) {
			i = check_valid_addresses( pid, l );
			if ( i ) {
				return;
			}
			i = check_valid_addresses( pid, l );
			if ( i ) {
				return;
			}
		}
		
		search_virtual_memory( pid, l, ll, pargv[optind+3] );
		break;

	case DISASS_VIRTUAL_MEMORY:
		if ( pargc != 5 ) {
			usage_virtm();
			return;
		}

		pid = atoi( pargv[optind] );
		XSTRTOUL( pargv[optind+1], l );

		if ( !force ) {
			i = check_valid_addresses( pid, l );
			if ( i ) {
				return;
			}
		}

		disassemble_virtual_memory( pid, l, atoi(pargv[optind+2]) );
		break;

	case EXE_VIRTUAL_MEMORY:
		if ( pargc != 4 ) {
			usage_virtm();
			return;
		}

		pid = atoi( pargv[optind] );
		get_exe_image( pid, pargv[optind+1] );
		break;
		
	default:
		return;
	}

	return;
}
